#include <shogun/base/init.h>
#include <shogun/base/some.h>
#include <shogun/lib/SGVector.h>
#include <shogun/machine/gp/GaussianLikelihood.h>
#include <shogun/io/SerializableAsciiFile.h>
#include <shogun/evaluation/MeanSquaredError.h>
#include <shogun/machine/gp/FITCInferenceMethod.h>
#include <shogun/kernel/GaussianKernel.h>
#include <shogun/machine/gp/ZeroMean.h>
#include <shogun/features/DenseFeatures.h>
#include <shogun/io/CSVFile.h>
#include <shogun/lib/WrappedObjectArray.h>
#include <shogun/labels/RegressionLabels.h>
#include <shogun/regression/GaussianProcessRegression.h>

using namespace shogun;

int main(int, char*[])
{
init_shogun_with_defaults();

auto f_feats_train = some<CCSVFile>("../../data/regression_1d_sinc_features_train.dat");
auto f_feats_inducing = some<CCSVFile>("../../data/regression_1d_sinc_features_train.dat");
auto f_feats_test = some<CCSVFile>("../../data/regression_1d_sinc_features_test.dat");
auto f_labels_train = some<CCSVFile>("../../data/regression_1d_sinc_labels_train.dat");
auto f_labels_test = some<CCSVFile>("../../data/regression_1d_sinc_labels_test.dat");

//![create_features]
auto features_train = some<CDenseFeatures<float64_t>>(f_feats_train);
auto features_test = some<CDenseFeatures<float64_t>>(f_feats_test);
auto labels_train = some<CRegressionLabels>(f_labels_train);
auto labels_test = some<CRegressionLabels>(f_labels_test);
//![create_features]

//![create_kernel_and_mean_function]
auto width = 1.0;
auto kernel = some<CGaussianKernel>(features_train, features_train, width);
auto mean_function = some<CZeroMean>();
//![create_kernel_and_mean_function]

//![create_inference]
auto inducing_points = SGVector<int32_t>(3);
inducing_points[0] = 0;
inducing_points[1] = 1;
inducing_points[2] = 2;
auto inducing_features = some<CDenseFeatures<float64_t>>(f_feats_inducing);
inducing_features->add_subset(inducing_points);

auto gauss_likelihood = some<CGaussianLikelihood>();
auto inference_method = some<CFITCInferenceMethod>(kernel, features_train, mean_function, labels_train, gauss_likelihood, inducing_features);
//![create_inference]

//![create_instance]
auto gp_regression = some<CGaussianProcessRegression>(inference_method);
//![create_instance]

//![train_and_apply]
gp_regression->train();
auto labels_predict = gp_regression->apply_regression(features_test);
//![train_and_apply]

//![compute_variance]
auto variances = gp_regression->get_variance_vector(features_test);
//![compute_variance]

//![evaluate_error]
auto mserror = some<CMeanSquaredError>();
auto mse = mserror->evaluate(labels_predict, labels_test);
//![evaluate_error]

// integration testing variables
auto output = labels_test->get_labels();

// Serialize output for integration testing (automatically generated)
auto __sg_storage = some<CWrappedObjectArray>();
auto __sg_storage_file = some<CSerializableAsciiFile>("sparse_gaussian_process_regression.dat", 'w');
__sg_storage->append_wrapped_real(width, "width");
__sg_storage->append_wrapped_real_vector(variances, "variances");
__sg_storage->append_wrapped_real(mse, "mse");
__sg_storage->append_wrapped_real_vector(output, "output");
__sg_storage->save_serializable(__sg_storage_file);

exit_shogun();
return 0;
}
